/****************************   LICENSE   *****************************

Copyright (C) 2019, HUSEYIN TOKPUNAR http://huseyintokpinar.com/

Download & Update Latest Version: https://github.com/hhtokpinar/sqfEntity

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


**********************************************************************/

import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'package:flutter/services.dart';
import 'package:path/path.dart';
import 'package:sqfentity_base/sqfentity_base.dart';
import 'package:sqflite/sqflite.dart';
import 'package:synchronized/synchronized.dart';

// BEGIN DATABASE PROVIDER

class SqfEntityProvider extends SqfEntityModelBase {
  SqfEntityProvider(dynamic dbModel, {String tableName, String colId}) {
    _tableName = tableName;
    _colId = colId;
    _databaseName = dbModel.databaseName as String;
    _bundledDatabasePath = dbModel.bundledDatabasePath as String;
  }
  SqfEntityProvider._internal();
  static final SqfEntityProvider _sqfEntityProvider =
      SqfEntityProvider._internal();
  static SqfEntityProvider get = _sqfEntityProvider;
  String _databaseName;
  String _bundledDatabasePath;
  String _tableName = '';
  String _colId = '';

  static Database _db;

  Future<Database> get db async {
    return _db = _db ?? await initializeDb();
  }

  /// When the software/app is started, you must check the database was it initialized. If needed, initilizeDb method runs that CREATE TABLE / ALTER TABLE ADD COLUMN queries for you.
  Future<Database> initializeDb() async {
    final lock = Lock();
    Database _db;

    if (_db == null) {
      await lock.synchronized(() async {
        if (_db == null) {
          final databasesPath = await getDatabasesPath();
          final path = join(databasesPath, _databaseName);
          final file = File(path);

          // check if file exists
          if (file.existsSync()) {
            // Copy from asset if MyDbModel.bundledDatabasePath is not empty
            if (_bundledDatabasePath != null && _bundledDatabasePath != '') {
              final ByteData data = await rootBundle.load(_bundledDatabasePath);
              final List<int> bytes = data.buffer
                  .asUint8List(data.offsetInBytes, data.lengthInBytes);
              await File(path).writeAsBytes(bytes).then((_) {
                print('$_databaseName copied from asset successfully');
              });
            }
          }
          _db = await openDatabase(path, version: 1, onCreate: _createDb);
        }
      });
    }
    return _db;
  }

  /// Creates db if not exist
  void _createDb(Database db, int version) async {
    await db.execute(
        'Create table sqfentitytables (id integer primary key, tablename text, version int, properties text)');
    await db.execute(
        'Create table sqfentitysequences (id text UNIQUE, value integer)');
    print('$_databaseName created successfully');
  }

  Future<dynamic> getById(dynamic id) async {
    final Database db = await this.db;
    final query = 'Select * from $_tableName where $_colId=?';
    final result = await db.rawQuery(query, [id]);
    return result;
  }

  /// Run sql command with arguments (arguments is optional)
  Future<BoolResult> execSQL(String pSql, [List<dynamic> arguments]) async {
    final BoolResult result = BoolResult();
    final Database db = await this.db;
    try {
      await db.execute(pSql, arguments);
      result.success = true;
      result.successMessage = 'sql command executed successfully';
    } catch (e) {
      result.errorMessage = e.toString();
    }

    return result;
  }

  /// Run sql command List
  Future<BoolResult> execSQLList(List<String> pSql) async {
    final Database db = await this.db;
    final result = BoolResult(success: false);
    await db.transaction((txn) {
      for (String sql in pSql) {
        try {
          txn.execute(sql).then((_) {
            pSql.remove(sql);
            if (pSql.isEmpty) {
              result.success = true;
              result.successMessage = 'sql command list executed successfuly';
              return result;
            }
            return null;
          });
        } catch (e) {
          result.errorMessage = e.toString();
        }
      }
      return;
    });
    return result;
  }

  /// Run Select Command and return List<Map<String,dynamic>> such as datatable
  Future<List> execDataTable(String pSql, [List<dynamic> arguments]) async {
    final Database db = await this.db;
    final result = await db.rawQuery(pSql);
    return result;
  }

  /// Run Select Command and return first col of first row
  Future<dynamic> execScalar(String pSql, [List<dynamic> arguments]) async {
    final Database db = await this.db;
    if (!pSql.contains(' LIMIT ')) {
      pSql += ' LIMIT 1';
    }
    final result = await db.rawQuery(pSql);
    if (result.isNotEmpty)
      return result.first.values.first;
    else
      return null;
  }

  Future<List> toList(QueryParams params) async {
    final Database db = await this.db;
    final result = await db.query(_tableName,
        columns: params.selectColumns,
        where: params.whereString,
        whereArgs: params.whereArguments,
        orderBy: params.orderBy == '' ? null : params.orderBy,
        groupBy: params.groupBy == '' ? null : params.groupBy,
        limit: params.limit == 0 ? null : params.limit,
        offset: params.offset == 0 ? null : params.offset,
        distinct: params.distinct);
    print('\r\n');
    print('\r\n');

    // You can uncomment following print command for print when called db query with parameters automatically
    /*
    print('********** SqfEntityProvider.toList(QueryParams=> columns:' +
        (params.selectColumns != null ? params.selectColumns.toString() : '*') +
        ', whereString: ' +
        (params.whereString != null ? params.whereString : 'null') +
        ', whereArgs:' +
        (params.whereArguments != null
            ? params.whereArguments.toString()
            : 'null') +
        ', orderBy:' +
        (params.orderBy != '' ? params.orderBy : 'null') +
        ', groupBy:' +
        (params.groupBy != '' ? params.groupBy : 'null') +
        ')');
        */
    return result;
  }

  Future<BoolResult> delete(QueryParams params) async {
    final result = BoolResult();
    final Database db = await this.db;
    try {
      final deletedItems = await db.delete(_tableName,
          where: params.whereString, whereArgs: params.whereArguments);
      result.successMessage = '$deletedItems items deleted';
      result.success = true;
    } catch (e) {
      result.errorMessage = e.toString();
    }
    return result;
  }

  Future<BoolResult> updateBatch(
      QueryParams params, Map<String, dynamic> values) async {
    final result = BoolResult();
    final Database db = await this.db;
    try {
      final updatedItems = await db.update(_tableName, values,
          where: params.whereString, whereArgs: params.whereArguments);
      result.successMessage = '$updatedItems items updated';
      result.success = true;
    } catch (e) {
      result.errorMessage = e.toString();
    }
    return result;
  }

  Future<int> update(dynamic T) async {
    final Database db = await this.db;
    final o = await T.toMap(forQuery: true);
    final result = await db.update(_tableName, o as Map<String, dynamic>,
        where: '$_colId = ?', whereArgs: [o[_colId]]);
    return result;
  }

  Future<int> insert(dynamic T) async {
    final Database db = await this.db;
    final result = await db.insert(
        _tableName, await T.toMap(forQuery: true) as Map<String, dynamic>);
    return result;
  }

  Future<int> rawInsert(String pSql, List<dynamic> params) async {
    final Database db = await this.db;
    final result = await db.rawInsert(pSql, params);
    return result;
  }

  Future<List<BoolResult>> rawInsertAll(
      String pSql, List<dynamic> params) async {
    final results = List<BoolResult>();
    final Database db = await this.db;
    for (var t in params) {
      final result = BoolResult();
      try {
        final res = await db.rawInsert(pSql, t.toArgs() as List<dynamic>);
        result.success = true;
        result.successMessage = '$_colId: $res upserted successfuly';
      } catch (e) {
        result.successMessage = null;
        result.errorMessage = e.toString();
      }

      results.add(result);
    }

    return results;
  }

  Future<List<BoolResult>> saveAll(String pSql, List T) async {
    final results = List<BoolResult>();
    final Database db = await this.db;
    for (var t in T) {
      final result = BoolResult(success: false);
      try {
        final o = await t.toMap(forQuery: true);
        if (o[_colId] != null && o[_colId] != 0) {
          final uresult = await db.rawInsert(pSql, t.toArgs() as List<dynamic>);
          if (uresult > 0) {
            result.successMessage =
                'id=${o[_colId].toString()} upserted successfuly';
          }
        } else {
          final iresult =
              await db.insert(_tableName, o as Map<String, dynamic>);
          if (iresult > 0) {
            result.successMessage =
                'id=${iresult.toString()} inserted  successfuly';
          }
        }
        result.success = true;
      } catch (e) {
        result.successMessage = null;
        result.errorMessage = e.toString();
      }
      results.add(result);
    }
    return results;
  }

  Future<int> sequence(SqfEntitySequenceBase seq, bool increase,
      {bool reset = false}) async {
    final Database db = await this.db;
    if (reset) {
      await db.execute(
          'UPDATE sqfentitysequences SET value=${seq.startWith} where id=?',
          [seq.sequenceName]);
      return seq.startWith;
    }
    final resCurrent = await db.rawQuery(
        'SELECT value from sqfentitysequences where id=?', [seq.sequenceName]);
    final currentVal = resCurrent.first.values.first as int;
    if (!increase) {
      return currentVal;
    } else {
      final nextVal = currentVal + seq.incrementBy;
      if (nextVal >= seq.minValue && nextVal <= seq.maxValue) {
        await db.execute(
            'UPDATE sqfentitysequences SET value=$nextVal where id=?',
            [seq.sequenceName]);
        return nextVal;
      } else if (seq.cycle) {
        return sequence(seq, true, reset: true);
      } else {
        throw Exception(
            'SQFENTITIY: sequence (${seq.sequenceName}) exceeds its minValue or maxValue ');
      }
    }
  }
}
// END DATABASE PROVIDER


abstract class SqfEntityModelProvider extends SqfEntityModelBase {
  
  Future<bool> initializeDB() async {
    databaseTables = databaseTables ?? [];
    sequences = sequences ?? [];
    final dbSequences = sequences.where((i) => !i.initialized).toList();
    if (dbSequences.isNotEmpty) {
      final tableSquence = await SqfEntityProvider(this)
          .execDataTable("PRAGMA table_info('sqfentitysequences')");
      if (tableSquence == null || tableSquence.isEmpty) {
        await SqfEntityProvider(this).execSQL(
            'Create table sqfentitysequences (id text UNIQUE, value integer)');
      }
      for (SqfEntitySequenceBase sequence in dbSequences) {
        final sqRow = await SqfEntityProvider(this).execDataTable(
            'SELECT * FROM sqfentitysequences WHERE id=?',
            [sequence.sequenceName]);
        if (sqRow.isEmpty) {
          await SqfEntityProvider(this).execSQL(
              'INSERT INTO sqfentitysequences (id, value) VALUES (?,?)',
              [sequence.sequenceName, sequence.startWith]);
        }
        sequence.initialized = true;
        print(
            'SQFENTITIY: Sequence [${sequence.sequenceName}] initialized successfuly');
      }
    }
    final dbTables = databaseTables.where((i) => !i.initialized).toList();
    if (dbTables.isNotEmpty) {
      //List<String> updateQueryList = List<String>();
      for (SqfEntityTableBase table in dbTables) {
        // check exiting table fields in the database
        final tableFields = await SqfEntityProvider(this)
            .execDataTable('PRAGMA table_info(${table.tableName})');
        final List<TableField> exitingDBfields = List<TableField>();
        if (tableFields != null && tableFields.isNotEmpty) {
          final String primaryKeyName = tableFields[0]['name'].toString();
          if (table.primaryKeyName != primaryKeyName) {
            throw Exception(
                'SQFENTITIY: DATABASE INITIALIZE ERROR The primary key name for table named [${table.tableName}] must be [$primaryKeyName]');
          }
          for (int i = 1; i < tableFields.length; i++) {
            exitingDBfields.add(TableField(tableFields[i]['name'].toString(),
                parseDbType(tableFields[i]['type'].toString())));
          }
          // create SQL Command for new columns
          final List<String> alterTableColsQuery =
              checkTableColumns(table, exitingDBfields);

          if (alterTableColsQuery.isNotEmpty) {
            print('SQFENTITIY: alterTableQuery => $alterTableColsQuery');

            final result =
                await SqfEntityProvider(this).execSQLList(alterTableColsQuery);
            if (result.success) {
              table.initialized = true;
              print(
                  'SQFENTITIY: Table named [${table.tableName}] was initialized successfuly (Added new columns)');
              if (checkForIsReadyDatabase(dbTables)) {
                return true;
              }
            }
          } else {
            table.initialized = true;
            print(
                'SQFENTITIY: Table named [${table.tableName}] was initialized successfuly (No added new columns)');
            if (checkForIsReadyDatabase(dbTables)) {
              return true;
            }
          }
        } else {
          // The table if not exist
          await SqfEntityProvider(this).execSQL(table.createTableSQL);
          final List<String> alterTableIndexesQuery = checkTableIndexes(table);
          table.initialized = true;
          print(
              'SQFENTITIY: Table named [${table.tableName}] was initialized successfuly with create new table');
          if (checkForIsReadyDatabase(dbTables)) {
            return true;
          }
          if (alterTableIndexesQuery.isNotEmpty) {
            print(
                'SQFENTITIY: alterTableIndexesQuery => $alterTableIndexesQuery');
            await SqfEntityProvider(this).execSQLList(alterTableIndexesQuery);
          }
        }
      }
    }
    return true;
  }

  /// create Model String and set the Clipboard (After debugging, press Ctrl+V to paste the model from the Clipboard)
  /// to call this method use SqfEntityDbModel.createSqfEntityModel

  /// Send any Object or Object List such as a product or productList to json String
  Future<String> toJson(dynamic T) async {
    if (T is List) {
      final list = List<dynamic>();
      for (var o in T) {
        list.add(await o.toMap());
      }
      return json.encode(list);
    } else {
      return json.encode(await T.toMap());
    }
  }

  /// Run sql command with arguments (arguments is optional)
  Future<BoolResult> execSQL(String sql, [List<dynamic> arguments]) async {
    return SqfEntityProvider(this).execSQL(sql, arguments);
  }

  /// Run sql command List
  Future<BoolResult> execSQLList(List<String> sql) async {
    return SqfEntityProvider(this).execSQLList(sql);
  }

  /// Run Select Command and return List<Map<String,dynamic>> such as datatable
  Future<List> execDataTable(String sql, [List<dynamic> arguments]) async {
    return SqfEntityProvider(this).execDataTable(sql, arguments);
  }

  /// Run Select Command and return first col of first row
  Future<dynamic> execScalar(String sql, [List<dynamic> arguments]) async {
    return SqfEntityProvider(this).execScalar(sql, arguments);
  }
}

/// check all tables is it ready
bool checkForIsReadyDatabase(List<SqfEntityTableBase> dbTables) {
  if (dbTables.where((i) => !i.initialized).isEmpty) {
    print('SQFENTITIY: The database is ready for use');
    return true;
  } else {
    return false;
  }
}

List<String> checkTableIndexes(SqfEntityTableBase table) {
  final alterTableQuery = List<String>();
  for (SqfEntityFieldType field in table.fields) {
    if (field is SqfEntityFieldRelationshipBase) {
      alterTableQuery.add(
          'CREATE INDEX IF NOT EXISTS IDX${field.relationshipName + field.fieldName} ON ${table.tableName} (${field.fieldName} ASC)');
    }
  }
  return alterTableQuery;
}

List<String> checkTableColumns(
    SqfEntityTableBase table, List<TableField> exitingDBfields) {
  final alterTableQuery = List<String>();
  for (var newField in table.fields) {
    final eField =
        exitingDBfields.where((x) => x.fieldName == newField.fieldName);
    if (eField.isNotEmpty) {
      // if (newField.dbType == DbType.bool) newField.dbType = DbType.numeric;
      if (!(newField.dbType == DbType.bool &&
              eField.toList()[0].fieldType == DbType.numeric) &&
          eField.toList()[0].fieldType != newField.dbType) {
        throw Exception(
            'SQFENTITIY DATABASE INITIALIZE ERROR: The type of column [${newField.fieldName}(${newField.dbType.toString()})] does not match the exiting column [${eField.toList()[0].fieldName}(${eField.toList()[0].fieldType.toString()})]');
      }
    } else {
      alterTableQuery.add(
          'ALTER TABLE ${table.tableName} ADD COLUMN ${newField.toSqLiteFieldString()}');
      if (newField.defaultValue!=null) {
        if (newField.dbType == DbType.text && newField.defaultValue.toString().isNotEmpty) {
          newField.defaultValue = "'${newField.defaultValue}'";
        }
        alterTableQuery.add(
            'UPDATE ${table.tableName} set ${newField.fieldName}=${newField.defaultValue}');
      }
      if (newField is SqfEntityFieldRelationshipBase) {
        alterTableQuery.add(
            'CREATE INDEX [IF NOT EXISTS] IDX${newField.relationshipName + newField.fieldName} ON ${table.tableName} (${newField.fieldName} [ASC])');
      }
    }
  }
  return alterTableQuery;
}
